import io
import logging
import numpy as np
import sys
import traceback

import gradio as gr

from PIL import Image
from gifr.common import init_logging, set_logging_level, create_parser, init_state, State, make_prediction
from gifr.colors import default_colors

PROG: str = "gifr-imgseg"

_logger = logging.getLogger(PROG)

state: State = None

PREDICTION_TYPE_AUTO = "auto"
PREDICTION_TYPE_BLUECHANNEL = "blue-channel"
PREDICTION_TYPE_GRAYSCALE = "grayscale"
PREDICTION_TYPE_INDEXEDPNG = "indexed-png"
PREDICTION_TYPES = [
    PREDICTION_TYPE_AUTO,
    PREDICTION_TYPE_BLUECHANNEL,
    PREDICTION_TYPE_GRAYSCALE,
    PREDICTION_TYPE_INDEXEDPNG,
]


def next_default_color(state: State):
    """
    Returns the next default color.

    :param state: the state
    :type state: State
    :return: the color tuple
    :rtype: tuple
    """
    if state.params["default_colors_index"] >= len(state.params["default_colors"]):
        state.params["default_colors_index"] = 0
    result = state.params["default_colors"][state.params["default_colors_index"]]
    state.params["default_colors_index"] += 1
    return result


def get_color(state: State, label: str):
    """
    Returns the color for the label.

    :param state: the state
    :type state: State
    :param label: the label to get the color for
    :type label: str
    :return: the RGB color tuple
    :rtype: tuple
    """
    if label not in state.params["colors"]:
        state.params["colors"][label] = next_default_color(state)
    return state.params["colors"][label]


def predict(img_file: str) -> np.ndarray:
    """
    Sends the image to the model and returns the result.

    :param img_file: the image to send
    :type img_file: str
    :return: the prediction result
    :rtype: str
    """
    global state
    state.logger.info("Loading: %s" % img_file)
    with open(img_file, "rb") as f:
        content = f.read()
    img = Image.open(img_file)

    data = make_prediction(state, content)
    if data is None:
        state.logger.error("No data received. Timeout or error?")
        return None

    # mask: num classes and turn into palette image
    mask = Image.open(io.BytesIO(data))
    pred_type = state.params["prediction_type"]
    if pred_type == PREDICTION_TYPE_AUTO:
        if mask.mode == "RGB":
            pred_type = PREDICTION_TYPE_BLUECHANNEL
        elif mask.mode == "L":
            pred_type = PREDICTION_TYPE_GRAYSCALE
        elif mask.mode == "P":
            pred_type = PREDICTION_TYPE_INDEXEDPNG
        else:
            raise Exception("Unhandled image mode: %s" % mask.mode)
        state.logger.info("Prediction type determined: %s" % pred_type)
    if pred_type == PREDICTION_TYPE_BLUECHANNEL:
        arr = np.asarray(mask)
        arr = arr[:, :, 2]  # blue channel
        num_classes = len(np.unique(arr)) - 1
        mask = Image.fromarray(arr, "P")
    elif pred_type == PREDICTION_TYPE_GRAYSCALE:
        arr = np.asarray(mask)
        num_classes = len(np.unique(arr)) - 1
        mask = Image.fromarray(arr, "P")
    elif pred_type == PREDICTION_TYPE_INDEXEDPNG:
        arr = np.asarray(mask)
        num_classes = len(np.unique(arr)) - 1
    else:
        raise Exception("Unhandled prediction type: %s" % state.params["prediction_type"])
    state.logger.info("# classes: %d" % num_classes)

    # new palette for mask
    palette = [0, 0, 0]  # background
    for i in range(num_classes):
        color = next_default_color(state)
        palette.extend(color)
    mask.putpalette(palette)
    state.logger.debug("palette: %s" % str(palette))

    # overlay mask
    mask = mask.convert("RGBA")
    if state.params["only_mask"]:
        combined = mask
    else:
        arr = np.asarray(mask)
        a = state.params["alpha"] * np.ones((arr.shape[0], arr.shape[1], 1), dtype=np.uint8)
        arr = np.concatenate([arr[:, :, 0:3], a], axis=2, dtype=np.uint8)
        mask = Image.fromarray(arr, "RGBA")
        combined = img.convert("RGBA")
        combined.paste(mask, (0, 0), mask=mask)

    return np.asarray(combined)


def create_interface(state: State) -> gr.Interface:
    """
    Generates the interface.
    """
    return gr.Interface(
        title=state.title,
        description=state.description,
        fn=predict,
        inputs=[
            gr.Image(type="filepath", label="Input"),
        ],
        outputs=[
            gr.Image(label="Prediction"),
        ],
        allow_flagging="never")


def post_init_state(state: State):
    """
    Finalizes the initialization of the state.

    :param state: the state to update
    :type state: State
    """
    state.logger = _logger
    state.params["colors"] = dict()
    state.params["default_colors"] = default_colors()
    state.params["default_colors_index"] = 0


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    global state
    init_logging()
    parser = create_parser("Image segmentation interface. Allows the user to select an image "
                           + "and display the generated pixel mask overlayed.",
                           PROG, model_channel_in="images", model_channel_out="predictions",
                           timeout=2.0, ui_title="Image segmentation",
                           ui_desc="Sends the selected image to the model and shows the result (overlay or pixel mask).")
    parser.add_argument("--prediction_type", choices=PREDICTION_TYPES, default=PREDICTION_TYPE_AUTO, help="The type of image that the model returns")
    parser.add_argument("--alpha", metavar="NUM", help="The alpha value to use for the overlay (0: transparent, 255: opaque).", default=128, type=int, required=False)
    parser.add_argument("--only_mask", action="store_true", help="Whether to show only the predicted mask rather than overlaying it.")
    parsed = parser.parse_args(args=args)
    set_logging_level(_logger, parsed.logging_level)
    state = init_state(parsed)
    post_init_state(state)
    ui = create_interface(state)
    ui.launch(show_api=False, share=parsed.share_interface, inbrowser=parsed.launch_browser)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(sys.argv[1:]), file=sys.stderr)
        return 1


if __name__ == '__main__':
    main()



import json
import logging
import sys
import traceback

import gradio as gr

from gifr.common import init_logging, set_logging_level, create_parser, init_state, State, make_prediction

PROG: str = "gifr-textgen"

_logger = logging.getLogger(PROG)

state: State = None


def predict(text: str) -> str:
    """
    Sends the text to the model and returns the completed text.

    :param text: the text to send
    :type text: str
    :return: the prediction result
    :rtype: str
    """
    global state
    state.logger.info("Completing: %s" % text)

    # build query
    d = {state.params["send_text"]: text}
    if state.params["history_on"]:
        if state.params["send_history"] is not None:
            d[state.params["send_history"]] = state.history
        if state.params["send_turns"] is not None:
            d[state.params["send_turns"]] = state.turns

    # perform query
    result = make_prediction(state, json.dumps(d))

    # parse response
    if result is None:
        result = "no result"
    else:
        if state.params["json_response"]:
            try:
                d = json.loads(result.decode())
                result = d[state.params["receive_prediction"]]
                if state.params["history_on"]:
                    if state.params["receive_history"] in d:
                        state.history = d[state.params["receive_history"]]
                        _logger.info("History: %s" % str(state.history))
                    if state.params["receive_turns"] in d:
                        state.turns = d[state.params["receive_turns"]]
                        _logger.info("Turns: %s" % str(state.turns))
            except:
                result = "Failed to parse: %s" % str(result)
        else:
            try:
                result = result.decode()
            except:
                result = "Failed to parse: %s" % str(result)

    state.logger.info("Prediction: %s" % result)
    if state.params["clean_response"]:
        result = result.strip()
        if result.endswith("</s>"):
            result = result[0:-4]
            result = result.strip()

    return result


def create_interface(state: State) -> gr.Interface:
    """
    Generates the interface.

    :param state: the state to use
    :type state: State
    """
    return gr.Interface(
        title=state.title,
        description=state.description,
        fn=predict,
        inputs=[
            gr.Textbox(label="Input"),
        ],
        outputs=[
            gr.Textbox(label="Prediction"),
        ],
        allow_flagging="never")


def post_init_state(state: State):
    """
    Finalizes the initialization of the state.

    :param state: the state to update
    :type state: State
    """
    state.logger = _logger
    state.history = ""
    state.turns = 0


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    global state
    init_logging()
    parser = create_parser("Text generation interface. Allows the user to enter text "
                           + "and display the text generated by the model.",
                           PROG, model_channel_in="text", model_channel_out="prediction",
                           timeout=1.0, ui_title="Text generation",
                           ui_desc="Sends the entered text to the model to complete and displays the result.")
    parser.add_argument("--send_text", metavar="FIELD", help="The field name in the JSON prompt used for sending the text, ignored if not provided.", default="prompt", type=str, required=False)
    parser.add_argument("--json_response", action="store_true", help="Whether the reponse is a JSON object.")
    parser.add_argument("--receive_prediction", metavar="FIELD", help="The field name in the JSON response used for receiving the predicted text, ignored if not provided.", default="text", type=str, required=False)
    parser.add_argument("--history_on", action="store_true", help="Whether to keep track of the interactions.")
    parser.add_argument("--send_history", metavar="FIELD", help="The field name in the JSON query to use for sending the input history, ignored if not provided.", default=None, type=str, required=False)
    parser.add_argument("--send_turns", metavar="FIELD", help="The field name in the JSON query to use for sending the number of turns in the interaction, ignored if not provided.", default=None, type=str, required=False)
    parser.add_argument("--receive_history", metavar="FIELD", help="The field name in the JSON response used for receiving the input history, ignored if not provided.", default=None, type=str, required=False)
    parser.add_argument("--receive_turns", metavar="FIELD", help="The field name in the JSON response used for receiving the number of turns in the interaction, ignored if not provided.", default=None, type=str, required=False)
    parser.add_argument("--clean_response", action="store_true", help="Whether to clean up the response.")
    parsed = parser.parse_args(args=args)
    set_logging_level(_logger, parsed.logging_level)
    state = init_state(parsed)
    post_init_state(state)
    ui = create_interface(state)
    ui.launch(show_api=False, share=parsed.share_interface, inbrowser=parsed.launch_browser)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(sys.argv[1:]), file=sys.stderr)
        return 1


if __name__ == '__main__':
    main()



import argparse
import logging
import os
import redis

from dataclasses import dataclass, field
from datetime import datetime
from time import sleep


LOGGING_DEBUG = "DEBUG"
LOGGING_INFO = "INFO"
LOGGING_WARN = "WARN"
LOGGING_ERROR = "ERROR"
LOGGING_CRITICAL = "CRITICAL"
LOGGING_LEVELS = [
    LOGGING_DEBUG,
    LOGGING_INFO,
    LOGGING_WARN,
    LOGGING_ERROR,
    LOGGING_CRITICAL,
]

ENV_GIFR_LOGLEVEL = "GIFR_LOGLEVEL"
""" environment variable for the global default logging level. """


@dataclass
class State:
    connection: redis.Redis = None
    pubsub = None
    channel_out: str = "redis_in"
    channel_in: str = "redis_out"
    timeout: float = 5.0
    title: str = "gifr"
    description: str = ""
    data = None
    logger: logging.Logger = None
    params: dict = field(default_factory=dict)


def str_to_logging_level(level: str) -> int:
    """
    Turns a logging level string into the corresponding integer constant.

    :param level: the level to convert
    :type level: str
    :return: the int level
    :rtype: int
    """
    if level not in LOGGING_LEVELS:
        raise Exception("Invalid logging level (%s): %s" % ("|".join(LOGGING_LEVELS), level))
    if level == LOGGING_CRITICAL:
        return logging.CRITICAL
    elif level == LOGGING_ERROR:
        return logging.ERROR
    elif level == LOGGING_WARN:
        return logging.WARN
    elif level == LOGGING_INFO:
        return logging.INFO
    elif level == LOGGING_DEBUG:
        return logging.DEBUG
    else:
        raise Exception("Unhandled logging level: %s" % level)


def init_logging():
    """
    Initializes the logging.
    """
    level = logging.WARNING
    if os.getenv(ENV_GIFR_LOGLEVEL) is not None:
        level = str_to_logging_level(os.getenv(ENV_GIFR_LOGLEVEL))
    logging.basicConfig(level=level)


def set_logging_level(logger: logging.Logger, level: str):
    """
    Sets the logging level of the logger.

    :param logger: the logger to update
    :type logger: logging.Logger
    :param level: the level string, see LOGGING_LEVELS
    :type level: str
    """
    logger.setLevel(str_to_logging_level(level))


def create_parser(description: str, prog: str, host: str = "localhost", port: int = 6379, db: int = 0,
                  model_channel_in: str = "model_channel_in", model_channel_out: str = "model_channel_out",
                  timeout: float = 5.0, ui_title: str = "gifr", ui_desc: str = "") -> argparse.ArgumentParser:
    """
    Creates a base parser with options for redis.

    :param description: the description of the program
    :type description: str
    :param prog: the command-line program
    :type prog: str
    :param host: the default redis host
    :type host: str
    :param port: the default redis port
    :type port: int
    :param db: the redis database to use
    :type db: int
    :param model_channel_in: the redis channel to send the data to for making predictions
    :type model_channel_in: str
    :param model_channel_out: the redis channel to receive the predictions on
    :type model_channel_out: str
    :param timeout: the number of seconds to wait for a prediction
    :type timeout: float
    :param ui_title: the title to use for the interface
    :type ui_title: str
    :param ui_desc: the description to use in the interface
    :type ui_desc: str
    """
    parser = argparse.ArgumentParser(
        description=description, prog=prog, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--redis_host", metavar="HOST", help="The host with the redis server.", default=host, type=str, required=False)
    parser.add_argument("--redis_port", metavar="PORT", help="The port of the redis server.", default=port, type=int, required=False)
    parser.add_argument("--redis_db", metavar="DB", help="The redis database to use.", default=db, type=int, required=False)
    parser.add_argument("--model_channel_in", metavar="CHANNEL", help="The channel to send the data to for making predictions.", default=model_channel_in, type=str, required=False)
    parser.add_argument("--model_channel_out", metavar="CHANNEL", help="The channel to receive the predictions on.", default=model_channel_out, type=str, required=False)
    parser.add_argument("--timeout", metavar="SECONDS", help="The number of seconds to wait for a prediction.", default=timeout, type=float, required=False)
    parser.add_argument("--title", metavar="TITLE", help="The title to use for interface.", default=ui_title, type=str, required=False)
    parser.add_argument("--description", metavar="DESC", help="The description to use in the interface.", default=ui_desc, type=str, required=False)
    parser.add_argument("--launch_browser", action="store_true", help="Whether to automatically launch the interface in a new tab of the default browser.")
    parser.add_argument("--share_interface", action="store_true", help="Whether to publicly share the interface at https://XYZ.gradio.live/.")
    parser.add_argument("--logging_level", choices=LOGGING_LEVELS, default=LOGGING_WARN, help="The logging level to use")
    return parser


def init_state(ns: argparse.Namespace) -> State:
    """
    Initializes the redis state container with the supplied parsed parameters.

    :param ns: the parsed options
    :type ns: argparse.Namespace
    :return: the state container
    :rtype: State
    """
    result = State(
        connection=redis.Redis(host=ns.redis_host, port=ns.redis_port, db=ns.redis_db),
        channel_in=ns.model_channel_in,
        channel_out=ns.model_channel_out,
        timeout=ns.timeout,
        title=ns.title,
        description=ns.description,
    )

    for att in dir(ns):
        if att.startswith("_"):
            continue
        if att in ["redis_host", "redis_port", "redis_db", "model_channel_in", "model_channel_out", "timeout", "title", "description"]:
            continue
        result.params[att] = getattr(ns, att)

    return result


def make_prediction(state: State, data):
    """
    Makes a prediction by broadcasting the data and waiting for a result coming through.

    :param state: the state to use to broadcasting/listening
    :type state: State
    :param data: the data to send
    :return: the received data, None if failed or timeout
    """
    def anon_handler(message):
        data = message['data']
        state.data = data
        state.pubsub_thread.stop()
        state.pubsub.close()
        state.pubsub = None

    state.pubsub = state.connection.pubsub()
    state.pubsub.psubscribe(**{state.channel_out: anon_handler})
    state.pubsub_thread = state.pubsub.run_in_thread(sleep_time=0.01)
    state.connection.publish(state.channel_in, data)

    # wait for data to show up
    start = datetime.now()
    end = start
    no_data = False
    while state.pubsub is not None:
        sleep(0.01)
        end = datetime.now()
        if state.timeout > 0:
            if (end - start).total_seconds() >= state.timeout:
                msg = "Timeout reached!"
                if state.logger is None:
                    print(msg)
                else:
                    state.logger.error(msg)
                no_data = True
                break

    if no_data:
        return None
    else:
        msg = "Time for prediction: %0.3f seconds" % (end - start).total_seconds()
        if state.logger is None:
            print(msg)
        else:
            state.logger.info(msg)
        return state.data



import json
import logging
import sys
import traceback

from typing import Tuple

import gradio as gr

from gifr.common import init_logging, set_logging_level, create_parser, init_state, State, make_prediction

PROG: str = "gifr-textclass"

_logger = logging.getLogger(PROG)

state: State = None


def predict(text: str) -> Tuple[str, float]:
    """
    Sends the text to the model and returns the label and score.

    :param text: the text to send
    :type text: str
    :return: the prediction result
    :rtype: str
    """
    global state
    state.logger.info("Classifying: %s" % text)
    d = {"text": text}
    prediction = make_prediction(state, json.dumps(d))
    if prediction is None:
        label = ""
        score = -1
    else:
        try:
            d = json.loads(prediction.decode())
            label = d["label"]
            score = d["score"]
        except:
            label = prediction.decode()
            score = -1

    state.logger.info("Prediction: %s -> label=%s, score=%f" % (prediction, label, score))
    return label, score


def create_interface(state: State) -> gr.Interface:
    """
    Generates the interface.

    :param state: the state to use
    :type state: State
    """
    return gr.Interface(
        title=state.title,
        description=state.description,
        fn=predict,
        inputs=[
            gr.Textbox(label="Input"),
        ],
        outputs=[
            gr.Textbox(label="Label"),
            gr.Textbox(label="Score"),
        ],
        allow_flagging="never")


def post_init_state(state: State):
    """
    Finalizes the initialization of the state.

    :param state: the state to update
    :type state: State
    """
    state.logger = _logger


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    global state
    init_logging()
    parser = create_parser("Text classification interface. Allows the user to enter text "
                           + "and display the predicted label and score returned by the model.",
                           PROG, model_channel_in="text", model_channel_out="prediction",
                           timeout=1.0, ui_title="Text classification",
                           ui_desc="Sends the entered text to the model to complete and displays the predicted label and score.")
    parsed = parser.parse_args(args=args)
    set_logging_level(_logger, parsed.logging_level)
    state = init_state(parsed)
    post_init_state(state)
    ui = create_interface(state)
    ui.launch(show_api=False, share=parsed.share_interface, inbrowser=parsed.launch_browser)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(sys.argv[1:]), file=sys.stderr)
        return 1


if __name__ == '__main__':
    main()



# taken from:
# https://github.com/waikato-ufdl/wai-annotations-imgvis/blob/6d7d29353311178a98cbd2df67e445a752245f8c/src/wai/annotations/imgvis/isp/annotation_overlay/component/_fonts.py

import traceback

from matplotlib import font_manager
from PIL import ImageFont


DEFAULT_FONT_FAMILY = "sans\\-serif"


def load_font(logger, family, size):
    """
    Attempts to instantiate the specified font family.

    :param logger: the logger instance to use, ignored if None
    :param family: the TTF font family
    :type family: str
    :param size: the size to use
    :type size: int
    :return: the Pillow font
    """
    try:
        mpl_font = font_manager.FontProperties(family=family)
        font_file = font_manager.findfont(mpl_font)
        return ImageFont.truetype(font_file, size)
    except:
        msg = "Failed to instantiate font family '%s', falling back on '%s'" % (family, DEFAULT_FONT_FAMILY)
        if logger is not None:
            logger.warning(msg, exc_info=True)
        else:
            print(msg)
            print(traceback.format_exc())

        mpl_font = font_manager.FontProperties(family=DEFAULT_FONT_FAMILY)
        font_file = font_manager.findfont(mpl_font)
        return ImageFont.truetype(font_file, size)



import json
import logging
import sys
import traceback

import gradio as gr

from gifr.common import init_logging, set_logging_level, create_parser, init_state, State, make_prediction

PROG: str = "gifr-imgcls"

_logger = logging.getLogger(PROG)

state: State = None


def predict(img_file: str) -> str:
    """
    Sends the image to the model and returns the result.

    :param img_file: the image to send
    :type img_file: str
    :return: the prediction result
    :rtype: str
    """
    global state
    state.logger.info("Loading: %s" % img_file)
    with open(img_file, "rb") as f:
        content = f.read()

    data = make_prediction(state, content)
    if data is None:
        result = {"no result": 0.0}
    else:
        result = json.loads(data.decode())
    state.logger.info("Prediction: %s" % result)
    return result


def create_interface(state: State) -> gr.Interface:
    """
    Generates the interface.

    :param state: the state to use
    :type state: State
    """
    return gr.Interface(
        title=state.title,
        description="Sends the selected image to the model and displays the generated prediction results.",
        fn=predict,
        inputs=[
            gr.Image(type="filepath", label="Input"),
        ],
        outputs=[
            gr.Label(label="Prediction"),
        ],
        allow_flagging="never")


def post_init_state(state: State):
    """
    Finalizes the initialization of the state.

    :param state: the state to update
    :type state: State
    """
    state.logger = _logger


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    global state
    init_logging()
    parser = create_parser("Image classification interface. Allows the user to select an image "
                           + "and display the probabilities per label that the model generated.",
                           PROG, model_channel_in="images", model_channel_out="predictions",
                           timeout=1.0, ui_title="Image classification",
                           ui_desc="Sends the selected image to the model and displays the generated prediction results.")
    parsed = parser.parse_args(args=args)
    set_logging_level(_logger, parsed.logging_level)
    state = init_state(parsed)
    post_init_state(state)
    ui = create_interface(state)
    ui.launch(show_api=False, share=parsed.share_interface, inbrowser=parsed.launch_browser)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(sys.argv[1:]), file=sys.stderr)
        return 1


if __name__ == '__main__':
    main()



import gradio as gr
import json
import logging
import numpy as np
import os
import sys
import traceback

from datetime import datetime
from PIL import Image, ImageDraw
from typing import Tuple

from opex import ObjectPredictions, BBox
from gifr.common import init_logging, set_logging_level, create_parser, init_state, State, make_prediction
from gifr.colors import default_colors, text_color
from gifr.fonts import load_font, DEFAULT_FONT_FAMILY


PROG: str = "gifr-objdet"

_logger = logging.getLogger(PROG)

state: State = None


def next_default_color(state: State):
    """
    Returns the next default color.

    :param state: the state
    :type state: State
    :return: the color tuple
    :rtype: tuple
    """
    if state.params["default_colors_index"] >= len(state.params["default_colors"]):
        state.params["default_colors_index"] = 0
    result = state.params["default_colors"][state.params["default_colors_index"]]
    state.params["default_colors_index"] += 1
    return result


def get_color(state: State, label: str):
    """
    Returns the color for the label.

    :param state: the state
    :type state: State
    :param label: the label to get the color for
    :type label: str
    :return: the RGB color tuple
    :rtype: tuple
    """
    if label not in state.params["colors"]:
        state.params["colors"][label] = next_default_color(state)
    return state.params["colors"][label]


def get_outline_color(state, label):
    """
    Generates the color for the outline.

    :param state: the state
    :type state: State
    :param label: the label to get the color for
    :type label: str
    :return: the RGBA color tuple
    :rtype: tuple
    """
    r, g, b = get_color(state, label)
    return r, g, b, state.params["outline_alpha"]


def get_fill_color(state, label):
    """
    Generates the color for the filling.

    :param state: the state
    :type state: State
    :param label: the label to get the color for
    :type label: str
    :return: the RGBA color tuple
    :rtype: tuple
    """
    r, g, b = get_color(state, label)
    return r, g, b, state.params["fill_alpha"]


def expand_label(state: State, label: str, score: float) -> str:
    """
    Expands the label text.

    :param state: the state
    :type state: State
    :param label: the current label
    :type label: str
    :param score: the prediction score
    :type score: float
    :return: the expanded label text
    :rtype: str
    """
    result = state.params["text_format"].replace("{label}", label)
    result = result.replace("{score}", ("%." + str(state.params["num_decimals"]) + "f") % float(score))
    return result


def text_coords(state: State, draw: ImageDraw, text: str, rect: BBox) -> Tuple[int, int, int, int]:
    """
    Determines the text coordinates in the image.

    :param state: the state
    :type state: State
    :param draw: the ImageDraw instance
    :type draw: ImageDraw
    :param text: the text to output
    :type text: str
    :param rect: the rectangle to use as reference
    :return: the x, y, w, h tuple
    :rtype: tuple
    """
    try:
        w, h = draw.textsize(text, font=state.params["font"])
    except:
        # newer versions of Pillow deprecated ImageDraw.textsize
        # https://levelup.gitconnected.com/how-to-properly-calculate-text-size-in-pil-images-17a2cc6f51fd
        ascent, descent = state.params["font"].getmetrics()
        w = state.params["font"].getmask(text).getbbox()[2]
        h = state.params["font"].getmask(text).getbbox()[3] + descent

    horizontal = state.params["horizontal"]
    vertical = state.params["vertical"]

    # x
    if horizontal == "L":
        x = rect.left
    elif horizontal == "C":
        x = rect.left + (rect.right - rect.left - w) // 2
    elif horizontal == "R":
        x = rect.right - w
    else:
        raise Exception("Unhandled horizontal text position: %s" % horizontal)

    # y
    if vertical == "T":
        y = rect.top
    elif vertical == "C":
        y = rect.top + (rect.bottom - rect.top - h) // 2
    elif vertical == "B":
        y = rect.bottom - h
    else:
        raise Exception("Unhandled horizontal text position: %s" % horizontal)

    return x, y, w, h


def predict(img_file: str) -> np.ndarray:
    """
    Sends the image to the model and returns the result.

    :param img_file: the image to send
    :type img_file: str
    :return: the prediction result
    :rtype: str
    """
    global state
    state.logger.info("Loading: %s" % img_file)
    with open(img_file, "rb") as f:
        content = f.read()
    img = Image.open(img_file)

    data = make_prediction(state, content)
    if data is None:
        preds_str = json.dumps({
            "timestamp": datetime.now().strftime("%Y%m%d_%H%M%S.%f"),
            "id": os.path.basename(img_file),
            "objects": []
        })
    else:
        preds_str = data.decode()
    state.logger.info("Prediction: %s" % preds_str)
    preds = ObjectPredictions.from_json_string(preds_str)
    overlay = Image.new('RGBA', img.size, (0, 0, 0, 0))
    draw = ImageDraw.Draw(overlay)
    for i, obj in enumerate(preds.objects):
        if state.params["vary_colors"]:
            color_label = "object-%d" % i
        else:
            color_label = obj.label

        # assemble polygon
        points = []
        if not state.params["force_bbox"]:
            d = obj.polygon.to_array_pair_polygon()
            for x, y in zip(d["x"], d["y"]):
                points.append((x, y))
        else:
            rect = obj.bbox
            points.append((rect.left, rect.top))
            points.append((rect.right, rect.top))
            points.append((rect.right, rect.bottom))
            points.append((rect.left, rect.bottom))
        if state.params["fill"]:
            draw.polygon(tuple(points), outline=get_outline_color(state, color_label), fill=get_fill_color(state, color_label), width=state.params["outline_thickness"])
        else:
            draw.polygon(tuple(points), outline=get_outline_color(state, color_label), width=state.params["outline_thickness"])

        # output text
        if len(state.params["text_format"]) > 0:
            text = expand_label(state, obj.label, obj.score)
            rect = obj.bbox
            x, y, w, h = text_coords(state, draw, text, rect)
            draw.rectangle((x, y, x+w, y+h), fill=get_outline_color(state, color_label))
            draw.text((x, y), text, font=state.params["font"], fill=text_color(get_color(state, color_label)))

    img.paste(overlay, (0, 0), mask=overlay)

    return np.asarray(img)


def create_interface(state: State) -> gr.Interface:
    """
    Generates the interface.

    :param state: the state to use
    :type state: State
    """
    return gr.Interface(
        title=state.title,
        description=state.description,
        fn=predict,
        inputs=[
            gr.Image(type="filepath", label="Input"),
        ],
        outputs=[
            gr.Image(label="Predictions"),
        ],
        allow_flagging="never")


def post_init_state(state: State):
    """
    Finalizes the initialization of the state.

    :param state: the state to update
    :type state: State
    """
    state.logger = _logger
    state.params["colors"] = dict()
    state.params["default_colors"] = default_colors()
    state.params["default_colors_index"] = 0
    state.params["font"] = load_font(_logger, state.params["font_family"], state.params["font_size"])
    anchors = state.params["text_placement"].split(",")
    state.params["vertical"] = anchors[0]
    state.params["horizontal"] = anchors[1]


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    global state
    init_logging()
    parser = create_parser("Object detection interface. Allows the user to select an image "
                           + "and overlay the predictions that the model generated.",
                           PROG, model_channel_in="images", model_channel_out="predictions",
                           timeout=1.0, ui_title="Object detection",
                           ui_desc="Sends the selected image to the model and overlays the predicted objects on it in the output.")
    parser.add_argument("--min_score", metavar="FLOAT", help="The minimum score a prediction must have (0-1).", default=0.0, type=float, required=False)
    parser.add_argument("--text_format", metavar="FORMAT", help="The format for the text, placeholders: {label}, {score}.", default="{label}", type=str, required=False)
    parser.add_argument("--text_placement", metavar="V,H", help="Comma-separated list of vertical (T=top, C=center, B=bottom) and horizontal (L=left, C=center, R=right) anchoring.", default="T,L", type=str, required=False)
    parser.add_argument("--font_family", metavar="NAME", help="The name of the font family.", default=DEFAULT_FONT_FAMILY, type=str, required=False)
    parser.add_argument("--font_size", metavar="SIZE", help="The size of the font.", default=14, type=int, required=False)
    parser.add_argument("--num_decimals", metavar="NUM", help="The number of decimals to use for the score.", default=3, type=int, required=False)
    parser.add_argument("--outline_thickness", metavar="NUM", help="The line thickness to use for the outline, <1 to turn off.", default=3, type=int, required=False)
    parser.add_argument("--outline_alpha", metavar="NUM", help="The alpha value to use for the outline (0: transparent, 255: opaque).", default=255, type=int, required=False)
    parser.add_argument("--fill", action="store_true", help="Whether to fill the bounding boxes/polygons", required=False)
    parser.add_argument("--fill_alpha", metavar="NUM", help="The alpha value to use for the filling (0: transparent, 255: opaque).", default=128, type=int, required=False)
    parser.add_argument("--vary_colors", action="store_true", help="Whether to vary the colors of the outline/filling regardless of label", required=False)
    parser.add_argument("--force_bbox", action="store_true", help="Whether to force a bounding box even if there is a polygon available", required=False)
    parsed = parser.parse_args(args=args)
    set_logging_level(_logger, parsed.logging_level)
    state = init_state(parsed)
    post_init_state(state)
    ui = create_interface(state)
    ui.launch(show_api=False, share=parsed.share_interface, inbrowser=parsed.launch_browser)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(sys.argv[1:]), file=sys.stderr)
        return 1


if __name__ == '__main__':
    main()
